import numpy as np
import matplotlib.pyplot as plt
import Losses
import activations
import back_prop
import dataset
import forward_model
import backward_model
import forward_prop
import back_prop
import optimization
import Metric
import matplotlib
import layers
import init
import Batch
import Mini_batch
import matplotlib.animation as animation
from dataset import dataset
from Metric import evaluation_metrics
import matplotlib; matplotlib.use("TkAgg")
from optimization import momentum ,ADAM
from matplotlib import style